@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013, 2014 Apple Inc. All rights reserved.
+ * Copyright (C) 2013-2021 Apple Inc. All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -23,18 +23,20 @@
  * THE POSSIBILITY OF SUCH DAMAGE.
  */
 
-#ifndef DFGDesiredWeakReferences_h
-#define DFGDesiredWeakReferences_h
-
-#include <wtf/HashSet.h>
+#pragma once
 
 #if ENABLE(DFG_JIT)
 
-namespace JSC { 
+#include "StructureID.h"
+#include "WriteBarrier.h"
+#include <wtf/FixedVector.h>
+#include <wtf/HashSet.h>
+
+namespace JSC {
 
 class CodeBlock;
 class JSCell;
-class SlotVisitor;
+class JSValue;
 class VM;
 
 namespace DFG {
@@ -48,19 +50,23 @@ public:
     ~DesiredWeakReferences();
 
     void addLazily(JSCell*);
+    void addLazily(JSValue);
     bool contains(JSCell*);
-    
+
     void reallyAdd(VM&, CommonData*);
-    
-    void visitChildren(SlotVisitor&);
+
+    void finalize();
+
+    template<typename Visitor> void visitChildren(Visitor&);
 
 private:
     CodeBlock* m_codeBlock;
-    HashSet<JSCell*> m_references;
+    UncheckedKeyHashSet<JSCell*> m_cells;
+    UncheckedKeyHashSet<StructureID> m_structures;
+    FixedVector<WriteBarrier<JSCell>> m_finalizedCells;
+    FixedVector<StructureID> m_finalizedStructures;
 };
 
 } } // namespace JSC::DFG
 
 #endif // ENABLE(DFG_JIT)
-
-#endif // DFGDesiredWeakReferences_h
